import { async } from '../scheduler/async';
import { isValidDate } from '../util/isDate';
import { Subscriber } from '../Subscriber';
export function delay(delay, scheduler = async) {
    const delayFor = isValidDate(delay) ? +delay - scheduler.now() : Math.abs(delay);
    return (source) => source.lift(new DelayOperator(delayFor, scheduler));
}
class DelayOperator {
    constructor(delay, scheduler) {
        this.delay = delay;
        this.scheduler = scheduler;
    }
    call(subscriber, source) {
        return source.subscribe(new DelaySubscriber(subscriber, this.delay, this.scheduler));
    }
}
class DelaySubscriber extends Subscriber {
    constructor(destination, delay, scheduler) {
        super(destination);
        this.destination = destination;
        this.delay = delay;
        this.scheduler = scheduler;
        this.queue = [];
        this.active = false;
    }
    static dispatch(state) {
        const source = state.source;
        const queue = source.queue;
        const scheduler = state.scheduler;
        const destination = state.destination;
        while (queue.length > 0 && queue[0].time - scheduler.now() <= 0) {
            destination.next(queue.shift().value);
        }
        if (queue.length > 0) {
            const delay = Math.max(0, queue[0].time - scheduler.now());
            this.schedule(state, delay);
        }
        else if (source.isStopped) {
            source.destination.complete();
            source.active = false;
        }
        else {
            this.unsubscribe();
            source.active = false;
        }
    }
    _schedule(scheduler) {
        this.active = true;
        const { destination } = this;
        destination.add(scheduler.schedule(DelaySubscriber.dispatch, this.delay, {
            source: this,
            destination,
            scheduler,
        }));
    }
    _next(value) {
        const scheduler = this.scheduler;
        const message = new DelayMessage(scheduler.now() + this.delay, value);
        this.queue.push(message);
        if (this.active === false) {
            this._schedule(scheduler);
        }
    }
    _error(err) {
        this.queue.length = 0;
        this.destination.error(err);
        this.unsubscribe();
    }
    _complete() {
        if (this.queue.length === 0) {
            this.destination.complete();
        }
        this.unsubscribe();
    }
}
class DelayMessage {
    constructor(time, value) {
        this.time = time;
        this.value = value;
    }
}
//# sourceMappingURL=delay.js.map